import logging


from os import path, makedirs
from traceback import format_exc
from time import gmtime, strftime


class LogHelper:
    """This class allows to manage logger.

       Sets console log or file log -> Choose the log level. Allows to write log on different log levels.
    """

    def __init__(self, log_dir_path, logger_name=None, filename=None, log_level=None, log_dir_name=None, enable_console_log_mode=False):
        self._filename = filename
        self._log_level = log_level
        self._log_dir_path = log_dir_path
        self._log_dir_name = log_dir_name
        self._enable_console_log_mode = enable_console_log_mode

        self._create_log_directory()
        self._set_logger_preferences()
        self.logger = logging.getLogger(logger_name)

    def _set_logger_preferences(self) -> None:
        """Logger initializing.

        Sets console log mode or file log mode -> Sets filename and log directory if file log mode enabled.

        @:returns: None
        @:rtype: NoneType
        """
        log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

        numeric_level = getattr(logging, self._log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % self._log_level)

        if self._enable_console_log_mode:
            logging.basicConfig(format=log_format, level=numeric_level)
        else:
            if self._filename:
                filename = '{}/{}_{}.log'.format(self._log_dir_path, self._filename, strftime("%Y-%m-%d-%H-%M", gmtime()))
            else:
                filename = '{}/{}_{}.log'.format(self._log_dir_path, 'log', strftime("%Y-%m-%d-%H-%M", gmtime()))
            logging.basicConfig(filename=filename, format=log_format, level=numeric_level)

    def write(self, lvl: int, msg: str) -> None:
        """Writes a log message.

        @:param lvl: log level
        @:param msg: message to write
        @:type lvl: int
        @:type msg: str

        Writes a log message with certain log level and message.

        @:returns: None
        @:rtype: NoneType
        """

        if (not msg) and (not lvl):
            msg = '{}'.format(format_exc())
        self.logger.log(lvl, msg)

    def _create_log_directory(self) -> None:
        """Creates log directory.

        @:returns: None
        @:rtype: NoneType
        """

        if not self._log_dir_name:
            self._log_dir_name = 'log'
        self._log_dir_path = path.join(self._log_dir_path, self._log_dir_name)

        if not path.exists(self._log_dir_path):
            makedirs(self._log_dir_path)



import requests
from abc import ABC, abstractmethod
from .user_agents import user_agents_list
import copy
import time
from .proxy_helper import ProxyHelper, BadProxyError, ProxySettings


class Downloader(ABC):

    def __init__(self, logger, proxy_settings=ProxySettings, use_session=False, attempts=20,
                 use_user_agents=True) -> None:
        self.logger = logger
        self.attempts = attempts
        self.user_agents_list = user_agents_list
        self.use_session = use_session
        self.use_user_agents = use_user_agents

        self.initialize_request_maker()
        self.proxy_auth = {}
        self.proxy_settings = proxy_settings
        self.proxy_helper = ProxyHelper(self.logger, self.proxy_settings)

    def get_proxies(self, proxy):
        """
        Creating proxy object from simple proxy string
        :param proxy: proxy string with {ip}:{port} pattern
        :type proxy: str
        :return: proxy object
        :rtype: dict
        """
        # get login and password if proxy required logining
        login = self.proxy_auth.get(proxy, {}).get('login', '')
        password = self.proxy_auth.get(proxy, {}).get('password', '')
        proxy_string = '{}' + f'://{login}:{password}@{proxy}'
        proxies = {'https': proxy_string.format('https'),
                   'http': proxy_string.format('http')}
        return proxies

    def create_request(self, method, url, params={}, cookies=None, data={}, headers={}, timeout=60, files=None):
        # Decorator must use instant function because we must change instant proxy dataframe only
        @self.proxy_helper.exception_decorator
        def request_to_page(proxy, **kwargs):
            """
            Request wrapper
            :param proxy: proxy string with {ip}:{port} pattern
            :type proxy: str
            :param kwargs: request params
            :return: response text
            :rtype: str
            """
            proxies = self.get_proxies(proxy) if proxy else {}
            start_time = time.time()
            response = self.request_maker.request(proxies=proxies, stream=True, **kwargs)
            full_content = b''
            for content in response.iter_content(1024):
                if time.time() - start_time > 30:
                    # if request time longer than 30 sec must stop request
                    raise BadProxyError
                full_content += content
            response._content = full_content

            return response

        headers = copy.deepcopy(headers)
        attempts = self.attempts
        while attempts > 0:
            if self.use_user_agents:
                random_agent = self.proxy_helper.get_user_agent()
                headers.update({'user-agent': random_agent})

            attempts -= 1
            # try without proxies last time
            raw_proxy = self.proxy_helper.get_proxy() if self.proxy_settings.use_proxy and attempts != 1 else None
            try:
                request_response = request_to_page(proxy=raw_proxy,
                                                   method=method,
                                                   url=url,
                                                   params=params,
                                                   cookies=cookies,
                                                   data=data,
                                                   headers=headers,
                                                   timeout=timeout,
                                                   files=files)

                return request_response
            except Exception as e:
                if isinstance(e, BadProxyError):
                    attempts += 1
                self.logger.write(10,
                                  'received {} exception on request on {} try on {} link'.format(e,
                                                                                                 self.attempts - attempts,
                                                                                                 url))

    def initialize_request_maker(self) -> None:
        """Session initializing.

        @:param self:
        @:type self: Downloader

        Enables session for requests if self.use_session else all requests without session.

        @:returns: None
        @:rtype: NoneType
        """

        if self.use_session:
            self.request_maker = requests.Session()
            # raise max pool size for use bigger thread pool size
            a = requests.adapters.HTTPAdapter(pool_maxsize=9999)
            self.request_maker.mount('https://', a)
        else:
            self.request_maker = requests

    def get(self, url, params={}, headers={}, cookies={}, timeout=60):
        """Makes 'GET' request with args by using create_request function.

        @:param self:
        @:param url: URL
        @:param params: params for 'GET' request
        @:param headers: request headers
        @:param cookies: request cookies
        @:param timeout: request timeout
        @:param return_response: enable to return response object
        @:param return_response_text: enable to return analog of response.text
        @:type self: Downloader
        @:type url: str
        @:type params: dict
        @:type headers: dict
        @:type cookies: dict
        @:type timeout: int

        Calls create_request function with args for 'GET' request.

        @:returns: tuple of response and response content
        @:rtype: tuple
        """
        return self.create_request(method='GET',
                                   url=url,
                                   params=params,
                                   cookies=cookies,
                                   headers=headers,
                                   timeout=timeout,
                                   )

    def post(self, url, data={}, headers={}, cookies={}, timeout=60):
        """Makes 'POST' request with args by using create_request function.

        @:param self:
        @:param url: URL
        @:param data: data for 'POST' request
        @:param headers: request headers
        @:param cookies: request cookies
        @:param timeout: request timeout
        @:param return_response: enable to return response object
        @:param return_response_text: enable to return analog of response.text
        @:type self: Downloader
        @:type url: str
        @:type data: dict
        @:type headers: dict
        @:type cookies: dict
        @:type timeout: int


        Calls create_request function with args for 'POST' request.

        @:returns: tuple of response and response content
        @:rtype: tuple
        """
        return self.create_request(method='POST',
                                   url=url,
                                   cookies=cookies,
                                   data=data,
                                   headers=headers,
                                   timeout=timeout,
                                   )



user_agents_list = ['Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:64.0) Gecko/20100101 Firefox/64.0',
'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML',
'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/532.5 (KHTML like Gecko) Chrome/4.0.249.0 Safari/532.5;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/532.9 (KHTML like Gecko) Chrome/5.0.310.0 Safari/532.9;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/534.7 (KHTML like Gecko) Chrome/7.0.514.0 Safari/534.7;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/534.14 (KHTML like Gecko) Chrome/9.0.601.0 Safari/534.14;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.14 (KHTML like Gecko) Chrome/10.0.601.0 Safari/534.14;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.20 (KHTML like Gecko) Chrome/11.0.672.2 Safari/534.20;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/534.27 (KHTML like Gecko) Chrome/12.0.712.0 Safari/534.27;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.1 (KHTML like Gecko) Chrome/13.0.782.24 Safari/535.1;;',
'Mozilla/5.0 (Windows NT 6.0) AppleWebKit/535.2 (KHTML',
'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.7 (KHTML like Gecko) Chrome/16.0.912.36 Safari/535.7;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.0 x64; en-US; rv:1.9pre) Gecko/2008072421 Minefield/3.0.2pre;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.0.10) Gecko/2009042316 Firefox/3.0.10;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-GB; rv:1.9.0.11) Gecko/2009060215 Firefox/3.0.11 (.NET CLR 3.5.30729);;',
'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.9.1.6) Gecko/20091201 Firefox/3.5.6 GTB5;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.1; tr; rv:1.9.2.8) Gecko/20100722 Firefox/3.6.8 ( .NET CLR 3.5.30729; .NET4.0E);;',
'Mozilla/5.0 (Windows NT 6.1; rv:2.0.1) Gecko/20100101 Firefox/4.0.1;;',
'Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.0.1) Gecko/20100101 Firefox/4.0.1;;',
'Mozilla/5.0 (Windows NT 5.1; rv:5.0) Gecko/20100101 Firefox/5.0;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:6.0a2) Gecko/20110622 Firefox/6.0a2;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:7.0.1) Gecko/20100101 Firefox/7.0.1;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b4pre) Gecko/20100815 Minefield/4.0b4pre;;',
'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0 );;',
'Mozilla/4.0 (compatible; MSIE 5.5; Windows 98; Win 9x 4.90);;',
'Mozilla/5.0 (Windows; U; Windows XP) Gecko MultiZilla/1.6.1.0a;;',
'Mozilla/2.02E (Win95; U);;',
'Mozilla/3.01Gold (Win95; I);;',
'Mozilla/4.8 [en] (Windows NT 5.1; U);;',
'Mozilla/5.0 (Windows; U; Win98; en-US; rv:1.4) Gecko Netscape/7.1 (ax);;',
'Opera/7.50 (Windows XP; U);;',
'Opera/7.50 (Windows ME; U) [en];;',
'Opera/7.51 (Windows NT 5.1; U) [en];;',
'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; en) Opera 8.0;;',
'Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.2b) Gecko/20021001 Phoenix/0.2;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.23) Gecko/20090825 SeaMonkey/1.1.18;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/527  (KHTML like Gecko Safari/419.3) Arora/0.6 (Change: );;',
'Mozilla/5.0 (Windows; U; ; en-NZ) AppleWebKit/527  (KHTML like Gecko Safari/419.3) Arora/0.8.0;;',
'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; Avant Browser; Avant Browser; .NET CLR 1.0.3705; .NET CLR 1.1.4322; Media Center PC 4.0; .NET CLR 2.0.50727; .NET CLR 3.0.04506.30);;',
'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.8 (KHTML like Gecko) Beamrise/17.2.0.9 Chrome/17.0.939.0 Safari/535.8;;',
'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/535.2 (KHTML',
'Mozilla/5.0 (Windows NT 6.2) AppleWebKit/536.3 (KHTML',
'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/536.6 (KHTML like Gecko) Chrome/20.0.1092.0 Safari/536.6;;',
'Mozilla/5.0 (Windows NT 6.2) AppleWebKit/536.6 (KHTML',
'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML like Gecko) Chrome/22.0.1207.1 Safari/537.1;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:10.0.1) Gecko/20100101 Firefox/10.0.1;;',
'Mozilla/5.0 (Windows NT 6.1; rv:12.0) Gecko/20120403211507 Firefox/12.0;;',
'Mozilla/5.0 (Windows NT 6.0; rv:14.0) Gecko/20100101 Firefox/14.0.1;;',
'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:15.0) Gecko/20120427 Firefox/15.0a1;;',
'Mozilla/5.0 (Windows NT 6.2; Win64; x64; rv:16.0) Gecko/16.0 Firefox/16.0;;',
'iTunes/9.0.2 (Windows; N);;',
'Mozilla/5.0 (compatible; Konqueror/4.5; Windows) KHTML/4.5.4 (like Gecko);;',
'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/533.1 (KHTML like Gecko) Maxthon/3.0.8.2 Safari/533.1;;',
'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1);;',
'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0);;',
'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; Trid,',
'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0);;',
'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; Trident/5.0);;',
'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0);;',
'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.2; Trident/5.0);;',
'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.2; WOW64; Trident/5.0);;',
'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0);;',
'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.1; Trident/6.0);;',
'Mozilla/5.0 (compatible; MSIE 10.6; Windows NT 6.1; Trident/5.0; InfoPath.2; SLCC1; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729; .NET CLR 2.0.50727) 3gpp-gba UNTRUSTED/1.0;;',
'Opera/9.25 (Windows NT 6.0; U; en);;',
'Opera/9.80 (Windows NT 5.2; U; en) Presto/2.2.15 Version/10.10;;',
'Opera/9.80 (Windows NT 5.1; U; ru) Presto/2.7.39 Version/11.00;;',
'Opera/9.80 (Windows NT 6.1; U; en) Presto/2.7.62 Version/11.01;;',
'Opera/9.80 (Windows NT 5.1; U; zh-tw) Presto/2.8.131 Version/11.10;;',
'Opera/9.80 (Windows NT 6.1; U; es-ES) Presto/2.9.181 Version/12.00;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/531.21.8 (KHTML like Gecko) Version/4.0.4 Safari/531.21.10;;',
'Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/533.17.8 (KHTML like Gecko) Version/5.0.1 Safari/533.17.8;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.19.4 (KHTML like Gecko) Version/5.0.2 Safari/533.18.5;;',
'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-GB; rv:1.9.1.17) Gecko/20110123 (like Firefox/3.x) SeaMonkey/2.0.12;;',
'Mozilla/5.0 (Windows NT 5.2; rv:10.0.1) Gecko/20100101 Firefox/10.0.1 SeaMonkey/2.7.1;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_5_8; en-US) AppleWebKit/532.8 (KHTML like Gecko) Chrome/4.0.302.2 Safari/532.8;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_4; en-US) AppleWebKit/534.3 (KHTML like Gecko) Chrome/6.0.464.0 Safari/534.3;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_5; en-US) AppleWebKit/534.13 (KHTML like Gecko) Chrome/9.0.597.15 Safari/534.13;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_2) AppleWebKit/535.1 (KHTML like Gecko) Chrome/14.0.835.186 Safari/535.1;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.2 (KHTML like Gecko) Chrome/15.0.874.54 Safari/535.2;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.7 (KHTML like Gecko) Chrome/16.0.912.36 Safari/535.7;;',
'Mozilla/5.0 (Macintosh; U; Mac OS X Mach-O; en-US; rv:2.0a) Gecko/20040614 Firefox/3.0.0 ;;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10.5; en-US; rv:1.9.0.3) Gecko/2008092414 Firefox/3.0.3;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; en-US; rv:1.9.1) Gecko/20090624 Firefox/3.5;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.6; en-US; rv:1.9.2.14) Gecko/20110218 AlexaToolbar/alxf-2.0 Firefox/3.6.14;;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10.5; en-US; rv:1.9.2.15) Gecko/20110303 Firefox/3.6.15;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:2.0.1) Gecko/20100101 Firefox/4.0.1;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:5.0) Gecko/20100101 Firefox/5.0;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:9.0) Gecko/20100101 Firefox/9.0;;',
'Mozilla/4.0 (compatible; MSIE 5.15; Mac_PowerPC);;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en-US) AppleWebKit/125.4 (KHTML like Gecko Safari) OmniWeb/v563.15;;',
'Opera/9.0 (Macintosh; PPC Mac OS X; U; en);;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en) AppleWebKit/125.2 (KHTML like Gecko) Safari/85.8;;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en) AppleWebKit/125.2 (KHTML like Gecko) Safari/125.8;;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; fr-fr) AppleWebKit/312.5 (KHTML like Gecko) Safari/312.3;;',
'Mozilla/5.0 (Macintosh; U; PPC Mac OS X; en) AppleWebKit/418.8 (KHTML like Gecko) Safari/419.3;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:2.0.1) Gecko/20100101 Firefox/4.0.1 Camino/2.2.1;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:2.0b6pre) Gecko/20100907 Firefox/4.0b6pre Camino/2.2a1pre;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_0) AppleWebKit/536.3 (KHTML like Gecko) Chrome/19.0.1063.0 Safari/536.3;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_2) AppleWebKit/537.4 (KHTML like Gecko) Chrome/22.0.1229.79 Safari/537.4;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_2; rv:10.0.1) Gecko/20100101 Firefox/10.0.1;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:16.0) Gecko/20120813 Firefox/16.0;;',
'iTunes/4.2 (Macintosh; U; PPC Mac OS X 10.2);;',
'iTunes/9.0.3 (Macintosh; U; Intel Mac OS X 10_6_2; en-ca);;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X; en-US) AppleWebKit/528.16 (KHTML like Gecko Safari/528.16) OmniWeb/v622.8.0.112941;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_5_6; en-US) AppleWebKit/528.16 (KHTML like Gecko Safari/528.16) OmniWeb/v622.8.0;;',
'Opera/9.20 (Macintosh; Intel Mac OS X; U; en);;',
'Opera/9.64 (Macintosh; PPC Mac OS X; U; en) Presto/2.1.1;;',
'Opera/9.80 (Macintosh; Intel Mac OS X; U; en) Presto/2.6.30 Version/10.61;;',
'Opera/9.80 (Macintosh; Intel Mac OS X 10.4.11; U; en) Presto/2.7.62 Version/11.00;;',
'Opera/9.80 (Macintosh; Intel Mac OS X 10.6.8; U; fr) Presto/2.9.168 Version/11.52;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_2; en-us) AppleWebKit/531.21.8 (KHTML like Gecko) Version/4.0.4 Safari/531.21.10;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_5; de-de) AppleWebKit/534.15  (KHTML like Gecko) Version/5.0.3 Safari/533.19.4;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; en-us) AppleWebKit/533.20.25 (KHTML like Gecko) Version/5.0.4 Safari/533.20.27;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_7; en-us) AppleWebKit/534.20.8 (KHTML like Gecko) Version/5.1 Safari/534.20.8;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_3) AppleWebKit/534.55.3 (KHTML like Gecko) Version/5.1.3 Safari/534.53.10;;',
'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.5; rv:10.0.1) Gecko/20100101 Firefox/10.0.1 SeaMonkey/2.7.1;;',
'ELinks (0.4pre5; Linux 2.6.10-ac7 i686; 80x33);;',
'ELinks/0.9.3 (textmode; Linux 2.6.9-kanotix-8 i686; 127x41);;',
'ELinks/0.12~pre5-4;;',
'Links/0.9.1 (Linux 2.4.24; i386;);;',
'Links (2.1pre15; Linux 2.4.26 i686; 158x61);;',
'Links (2.3pre1; Linux 2.6.38-8-generic x86_64; 170x48);;',
'w3m/0.5.1;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US) AppleWebKit/532.4 (KHTML like Gecko) Chrome/4.0.237.0 Safari/532.4 Debian;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US) AppleWebKit/532.8 (KHTML like Gecko) Chrome/4.0.277.0 Safari/532.8;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/532.9 (KHTML like Gecko) Chrome/5.0.309.0 Safari/532.9;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/534.7 (KHTML like Gecko) Chrome/7.0.514.0 Safari/534.7;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/540.0 (KHTML like Gecko) Ubuntu/10.10 Chrome/9.1.0.0 Safari/540.0;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/534.15 (KHTML like Gecko) Chrome/10.0.613.0 Safari/534.15;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US) AppleWebKit/534.15 (KHTML like Gecko) Ubuntu/10.10 Chromium/10.0.613.0 Chrome/10.0.613.0 Safari/534.15;;',
'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/534.24 (KHTML like Gecko) Ubuntu/10.10 Chromium/12.0.703.0 Chrome/12.0.703.0 Safari/534.24;;',
'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/535.1 (KHTML like Gecko) Chrome/13.0.782.20 Safari/535.1;;',
'Mozilla/5.0 Slackware/13.37 (X11; U; Linux x86_64; en-US) AppleWebKit/535.1 (KHTML like Gecko) Chrome/13.0.782.41;;',
'Mozilla/5.0 (X11; Linux i686) AppleWebKit/535.1 (KHTML like Gecko) Ubuntu/11.04 Chromium/14.0.825.0 Chrome/14.0.825.0 Safari/535.1;;',
'Mozilla/5.0 (X11; Linux i686) AppleWebKit/535.2 (KHTML like Gecko) Ubuntu/11.10 Chromium/15.0.874.120 Chrome/15.0.874.120 Safari/535.2;;',
'Mozilla/5.0 (X11; U; Linux; i686; en-US; rv:1.6) Gecko Epiphany/1.2.5;;',
'Mozilla/5.0 (X11; U; Linux i586; en-US; rv:1.7.3) Gecko/20040924 Epiphany/1.4.4 (Ubuntu);;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.6) Gecko/20040614 Firefox/0.8;;',
'Mozilla/5.0 (X11; U; Linux x86_64; sv-SE; rv:1.8.1.12) Gecko/20080207 Ubuntu/7.10 (gutsy) Firefox/2.0.0.12;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.11) Gecko/2009060309 Ubuntu/9.10 (karmic) Firefox/3.0.11;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.1.2) Gecko/20090803 Ubuntu/9.04 (jaunty) Shiretoko/3.5.2;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.5) Gecko/20091107 Firefox/3.5.5;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.3) Gecko/20091020 Linux Mint/8 (Helena) Firefox/3.5.3;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.2.9) Gecko/20100915 Gentoo Firefox/3.6.9;;',
'Mozilla/5.0 (X11; U; Linux i686; pl-PL; rv:1.9.0.2) Gecko/20121223 Ubuntu/9.25 (jaunty) Firefox/3.8;;',
'Mozilla/5.0 (X11; Linux i686; rv:2.0b6pre) Gecko/20100907 Firefox/4.0b6pre;;',
'Mozilla/5.0 (X11; Linux i686 on x86_64; rv:2.0.1) Gecko/20100101 Firefox/4.0.1;;',
'Mozilla/5.0 (X11; Linux i686; rv:2.0.1) Gecko/20100101 Firefox/4.0.1;;',
'Mozilla/5.0 (X11; Linux x86_64; rv:2.0.1) Gecko/20100101 Firefox/4.0.1;;',
'Mozilla/5.0 (X11; Linux x86_64; rv:2.2a1pre) Gecko/20100101 Firefox/4.2a1pre;;',
'Mozilla/5.0 (X11; Linux i686; rv:5.0) Gecko/20100101 Firefox/5.0;;',
'Mozilla/5.0 (X11; Linux i686; rv:6.0) Gecko/20100101 Firefox/6.0;;',
'Mozilla/5.0 (X11; Linux x86_64; rv:7.0a1) Gecko/20110623 Firefox/7.0a1;;',
'Mozilla/5.0 (X11; Linux i686; rv:8.0) Gecko/20100101 Firefox/8.0;;',
'Mozilla/5.0 (X11; Linux x86_64; rv:10.0.1) Gecko/20100101 Firefox/10.0.1;;',
'Mozilla/5.0 (X11; U; Linux; i686; en-US; rv:1.6) Gecko Galeon/1.3.14;;',
'Mozilla/5.0 (X11; U; Linux ppc; en-US; rv:1.8.1.13) Gecko/20080313 Iceape/1.1.9 (Debian-1.1.9-5);;',
'Mozilla/5.0 (X11; U; Linux i686; pt-PT; rv:1.9.2.3) Gecko/20100402 Iceweasel/3.6.3 (like Firefox/3.6.3) GTB7.0;;',
'Mozilla/5.0 (X11; Linux x86_64; rv:5.0) Gecko/20100101 Firefox/5.0 Iceweasel/5.0;;',
'Mozilla/5.0 (X11; Linux i686; rv:6.0a2) Gecko/20110615 Firefox/6.0a2 Iceweasel/6.0a2;;',
'Konqueror/3.0-rc4; (Konqueror/3.0-rc4; i686 Linux;;datecode);;',
'Mozilla/5.0 (compatible; Konqueror/3.3; Linux 2.6.8-gentoo-r3; X11;;;',
'Mozilla/5.0 (compatible; Konqueror/3.5; Linux 2.6.30-7.dmz.1-liquorix-686; X11) KHTML/3.5.10 (like Gecko) (Debian package 4:3.5.10.dfsg.1-1 b1);;',
'Mozilla/5.0 (compatible; Konqueror/3.5; Linux; en_US) KHTML/3.5.6 (like Gecko) (Kubuntu);;',
'Mozilla/5.0 (X11; Linux x86_64; en-US; rv:2.0b2pre) Gecko/20100712 Minefield/4.0b2pre;;',
'Mozilla/5.0 (X11; U; Linux; i686; en-US; rv:1.6) Gecko Debian/1.6-7;;',
'MSIE (MSIE 6.0; X11; Linux; i686) Opera 7.23;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.1) Gecko/20061024 Firefox/2.0 (Swiftfox);;',
'Mozilla/5.0 (X11; U; Linux; en-US) AppleWebKit/527  (KHTML like Gecko Safari/419.3) Arora/0.10.1;;',
'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/536.5 (KHTML like Gecko) Chrome/19.0.1084.9 Safari/536.5;;',
'Mozilla/5.0 (X11; CrOS i686 2268.111.0) AppleWebKit/536.11 (KHTML like Gecko) Chrome/20.0.1132.57 Safari/536.11;;',
'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.4 (KHTML like Gecko) Chrome/22.0.1229.56 Safari/537.4;;',
'Mozilla/4.0 (compatible; Dillo 3.0);;',
'Mozilla/5.0 (X11; U; Linux i686; en-us) AppleWebKit/528.5  (KHTML like Gecko Safari/528.5 ) lt-GtkLauncher;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.1.16) Gecko/20120421 Gecko Firefox/11.0;;',
'Mozilla/5.0 (X11; Linux i686; rv:12.0) Gecko/20100101 Firefox/12.0 ;;',
'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:14.0) Gecko/20100101 Firefox/14.0.1;;',
'Mozilla/5.0 (X11; Linux i686; rv:16.0) Gecko/20100101 Firefox/16.0;;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.8) Gecko Galeon/2.0.6 (Ubuntu 2.0.6-2);;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.1.16) Gecko/20080716 (Gentoo) Galeon/2.0.6;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.13) Gecko/20100916 Iceape/2.0.8;;',
'Mozilla/5.0 (X11; Linux i686; rv:14.0) Gecko/20100101 Firefox/14.0.1 Iceweasel/14.0.1;;',
'Mozilla/5.0 (X11; Linux x86_64; rv:15.0) Gecko/20120724 Debian Iceweasel/15.02;;',
'Mozilla/5.0 (compatible; Konqueror/4.2; Linux) KHTML/4.2.4 (like Gecko) Slackware/13.0;;',
'Mozilla/5.0 (compatible; Konqueror/4.3; Linux) KHTML/4.3.1 (like Gecko) Fedora/4.3.1-3.fc11;;',
'Mozilla/5.0 (compatible; Konqueror/4.4; Linux) KHTML/4.4.1 (like Gecko) Fedora/4.4.1-1.fc12;;',
'Mozilla/5.0 (compatible; Konqueror/4.4; Linux 2.6.32-22-generic; X11; en_US) KHTML/4.4.3 (like Gecko) Kubuntu;;',
'Midori/0.1.10 (X11; Linux i686; U; en-us) WebKit/(531).(2) ;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.0.3) Gecko/2008092814 (Debian-3.0.1-1);;',
'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9a3pre) Gecko/20070330;;',
'Opera/9.64 (X11; Linux i686; U; Linux Mint; nb) Presto/2.1.1;;',
'Opera/9.80 (X11; Linux i686; U; en) Presto/2.2.15 Version/10.10;;',
'Opera/9.80 (X11; Linux x86_64; U; pl) Presto/2.7.62 Version/11.00;;',
'Mozilla/5.0 (X11; Linux i686) AppleWebKit/534.34 (KHTML like Gecko) QupZilla/1.2.0 Safari/534.34;;',
'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.17) Gecko/20110123 SeaMonkey/2.0.12;;',
'Mozilla/5.0 (X11; Linux i686; rv:10.0.1) Gecko/20100101 Firefox/10.0.1 SeaMonkey/2.7.1;;',
'Mozilla/5.0 (X11; U; Linux x86_64; us; rv:1.9.1.19) Gecko/20110430 shadowfox/7.0 (like Firefox/7.0;;',
'Mozilla/5.0 (X11; U; Linux i686; it; rv:1.9.2.3) Gecko/20100406 Firefox/3.6.3 (Swiftfox);;',
'Uzbl (Webkit 1.3) (Linux i686 [i686]);;',
'ELinks (0.4.3; NetBSD 3.0.2PATCH sparc64; 141x19);;',
'Links (2.1pre15; FreeBSD 5.3-RELEASE i386; 196x84);;',
'w3m/0.5.1;;',
'Mozilla/5.0 (X11; U; FreeBSD i386; en-US) AppleWebKit/532.0 (KHTML like Gecko) Chrome/4.0.207.0 Safari/532.0;;',
'Mozilla/5.0 (X11; U; OpenBSD i386; en-US) AppleWebKit/533.3 (KHTML like Gecko) Chrome/5.0.359.0 Safari/533.3;;',
'Mozilla/5.0 (X11; U; FreeBSD x86_64; en-US) AppleWebKit/534.16 (KHTML like Gecko) Chrome/10.0.648.204 Safari/534.16;;',
'Mozilla/5.0 (X11; U; SunOS sun4m; en-US; rv:1.4b) Gecko/20030517 Mozilla Firebird/0.6;;',
'Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.6) Gecko/20040406 Galeon/1.3.15;;',
'Mozilla/5.0 (compatible; Konqueror/3.5; NetBSD 4.0_RC3; X11) KHTML/3.5.7 (like Gecko);;',
'Mozilla/5.0 (compatible; Konqueror/3.5; SunOS) KHTML/3.5.1 (like Gecko);;',
'Mozilla/5.0 (X11; U; FreeBSD; i386; en-US; rv:1.7) Gecko;;',
'Mozilla/4.77 [en] (X11; I; IRIX;64 6.5 IP30);;',
'Mozilla/4.8 [en] (X11; U; SunOS; 5.7 sun4u);;',
'Mozilla/5.0 (Unknown; U; UNIX BSD/SYSV system; C -) AppleWebKit/527  (KHTML like Gecko Safari/419.3) Arora/0.10.2;;',
'Mozilla/5.0 (X11; FreeBSD amd64) AppleWebKit/536.5 (KHTML like Gecko) Chrome/19.0.1084.56 Safari/536.5;;',
'Mozilla/5.0 (X11; FreeBSD amd64) AppleWebKit/537.4 (KHTML like Gecko) Chrome/22.0.1229.79 Safari/537.4;;',
'Mozilla/5.0 (X11; U; OpenBSD arm; en-us) AppleWebKit/531.2  (KHTML like Gecko) Safari/531.2  Epiphany/2.30.0;;',
'Mozilla/5.0 (X11; U; FreeBSD amd64; en-us) AppleWebKit/531.2  (KHTML like Gecko) Safari/531.2  Epiphany/2.30.0;;',
'Mozilla/5.0 (X11; U; SunOS i86pc; en-US; rv:1.9.1b3) Gecko/20090429 Firefox/3.1b3;;',
'Mozilla/5.0 (X11; U; OpenBSD i386; en-US; rv:1.9.1) Gecko/20090702 Firefox/3.5;;',
'Mozilla/5.0 (X11; U; FreeBSD i386; de-CH; rv:1.9.2.8) Gecko/20100729 Firefox/3.6.8;;',
'Mozilla/5.0 (X11; FreeBSD amd64; rv:5.0) Gecko/20100101 Firefox/5.0;;',
'Mozilla/5.0 (compatible; Konqueror/4.1; DragonFly) KHTML/4.1.4 (like Gecko);;',
'Mozilla/5.0 (compatible; Konqueror/4.1; OpenBSD) KHTML/4.1.4 (like Gecko);;',
'Mozilla/5.0 (compatible; Konqueror/4.5; NetBSD 5.0.2; X11; amd64; en_US) KHTML/4.5.4 (like Gecko);;',
'Mozilla/5.0 (compatible; Konqueror/4.5; FreeBSD) KHTML/4.5.4 (like Gecko);;',
'Mozilla/5.0 (X11; U; NetBSD amd64; en-US; rv:1.9.2.15) Gecko/20110308 Namoroka/3.6.15;;',
'NetSurf/1.2 (NetBSD; amd64);;',
'Opera/9.80 (X11; FreeBSD 8.1-RELEASE i386; Edition Next) Presto/2.12.388 Version/12.10;;',
'Mozilla/5.0 (X11; U; SunOS i86pc; en-US; rv:1.8.1.12) Gecko/20080303 SeaMonkey/1.1.8;;',
'Mozilla/5.0 (X11; U; Linux armv6l; rv 1.8.1.5pre) Gecko/20070619 Minimo/0.020;;',
'Mozilla/5.0 (iPad; U; CPU OS 4_2_1 like Mac OS X; ja-jp) AppleWebKit/533.17.9 (KHTML like Gecko) Version/5.0.2 Mobile/8C148 Safari/6533.18.5;;',
'Mozilla/5.0 (iPad; CPU OS 6_0 like Mac OS X) AppleWebKit/536.26 (KHTML like Gecko) Version/6.0 Mobile/10A5355d Safari/8536.25;;',
'Mozilla/5.0 (Linux; U; Android 3.0.1; fr-fr; A500 Build/HRI66) AppleWebKit/534.13 (KHTML like Gecko) Version/4.0 Safari/534.13;;',
'Mozilla/5.0 (iPad; U; CPU OS 3_2 like Mac OS X; en-us) AppleWebKit/531.21.10 (KHTML like Gecko) Version/4.0.4 Mobile/7B334b Safari/531.21.10;;',
'Mozilla/5.0 (iPad; U; CPU OS 4_2_1 like Mac OS X; ja-jp) AppleWebKit/533.17.9 (KHTML like Gecko) Version/5.0.2 Mobile/8C148 Safari/6533.18.5;;',
'Mozilla/5.0 (iPad; U; CPU OS 4_3 like Mac OS X; en-us) AppleWebKit/533.17.9 (KHTML like Gecko) Version/5.0.2 Mobile/8F190 Safari/6533.18.5;;',
'Mozilla/5.0 (iPad; CPU OS 6_0 like Mac OS X) AppleWebKit/536.26 (KHTML like Gecko) Version/6.0 Mobile/10A5355d Safari/8536.25;;',
'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_5_7;en-us) AppleWebKit/530.17 (KHTML like Gecko) Version/4.0 Safari/530.17;;',
'Mozilla/5.0 (hp-tablet; Linux; hpwOS/3.0.2; U; de-DE) AppleWebKit/534.6 (KHTML like Gecko) wOSBrowser/234.40.1 Safari/534.6 TouchPad/1.0;;',
'Mozilla/5.0 (Linux; U; Android 3.0.1; en-us; GT-P7100 Build/HRI83) AppleWebkit/534.13 (KHTML like Gecko) Version/4.0 Safari/534.13;;',
'Mozilla/5.0 (Linux; U; Android 3.0.1; fr-fr; A500 Build/HRI66) AppleWebKit/534.13 (KHTML like Gecko) Version/4.0 Safari/534.13;;',
'Mozilla/5.0 (iPad; U; CPU OS 3_2 like Mac OS X; en-us) AppleWebKit/531.21.10 (KHTML like Gecko) Version/4.0.4 Mobile/7B334b Safari/531.21.10;;',
'Mozilla/5.0 (iPad; U; CPU OS 4_2_1 like Mac OS X; ja-jp) AppleWebKit/533.17.9 (KHTML like Gecko) Version/5.0.2 Mobile/8C148 Safari/6533.18.5;;',
'Mozilla/5.0 (iPad; U; CPU OS 4_3 like Mac OS X; en-us) AppleWebKit/533.17.9 (KHTML like Gecko) Version/5.0.2 Mobile/8F190 Safari/6533.18.5;;']



from urllib import parse
from traceback import format_exc
from .log_helper import LogHelper
from typing import Any, Dict, List
from contextlib import contextmanager
from sqlalchemy import text, create_engine, exc


class DatabaseHelper:
    """This class allows to manage database queries."""

    def __init__(self, db_credentials, log_dir_path) -> None:
        self.db_credentials = db_credentials
        self.logger = LogHelper(log_dir_path, logger_name='db_helper', log_level='INFO')

    def select(self, required_columns, filter_parameters={}, fetchsize=1000, query_string='', table_name='') -> Any:
        """Makes SELECT queries to get data from the database.

        @:param self:
        @:param required_columns: columns to get from db
        @:param filter_parameters: filters for WHERE query
        @:param fetchsize: max amount of records which returns from the database
        @:param query_string: custom query string
        @:param table_name: destination table name
        @:type self: DatabaseHelper
        @:type required_columns: list
        @:type filter_parameters: dict
        @:type fetchsize: int
        @:type query_string: str
        @:type table_name: str

        Makes SELECT queries, gets max 1k records from fetchmany in a loop.

        @:returns: dictionaries list if exists, else empty list, exception - False
        @:rtype: list or bool
        """

        with self.db_connection() as connection:
            try:
                if not query_string:
                    if not required_columns:
                        required_columns = '*'
                    else:
                        required_columns = ','.join([required_column for required_column in required_columns])

                    query_string = 'SELECT {} FROM {}'.format(required_columns, table_name)

                    if filter_parameters:
                        query_string += ' WHERE'

                        filter_parameters_counter = 1
                        filters_len = len(filter_parameters)

                        for filter_name, filter_parameter in filter_parameters.items():
                            if isinstance(filter_parameter, list):
                                filter_params = '(' + ','.join(["N'" + str(value).replace("'", "''") + "'" for value in
                                                                filter_parameter]) + ')'
                                column_match = ' {} IN {}'.format(filter_name, filter_params)
                            else:
                                column_match = ' {}={}'.format(filter_name,
                                                               "N'" + str(filter_parameter).replace("'", "''") + "'")
                            if filter_parameters_counter != filters_len:
                                column_match += ' AND'
                            query_string += column_match
                            filter_parameters_counter += 1

                attempts = 3
                while attempts:
                    try:
                        query_result = connection.execute(text(query_string))
                        break
                    except exc.OperationalError:
                        attempts -= 1

                columns = [column[0] for column in query_result.cursor.description]

                result_data = []
                for batch_result in self.gen_batch_result(query_result, columns, fetchsize):
                    result_data.extend(batch_result)
                return result_data
            except:
                self.logger.write(40, 'Can\'t read data from the table. - \n\n{}'.format(format_exc()))
                return False

    def insert(self, data, check_dup=False, col_to_check: str = '', query_string: str = '',
               table_name: str = '') -> bool:
        """Makes INSERT query to put data to the database.

        @:param self:
        @:param data: dict of dicts or list of dicts
        @:param check_dup: check or not for duplicates in db (PRIMARY KEY)
        @:param col_to_check: column to check for duplicate
        @:param query_string: custom query string
        @:param table_name: destination table name
        @:type self: DatabaseHelper
        @:type data: dict or list
        @:type check_dup: bool
        @:type col_to_check: str
        @:type query_string: str
        @:type table_name: str

        Makes SELECT query for each col_to_check from data parameter to detect duplicate col_to_check in the database ->
        makes INSERT query for unique col_to_check data.

        @:returns: True or False
        @:rtype: bool
        """

        with self.db_connection() as connection:
            try:
                if not query_string:
                    if check_dup:
                        data_to_write = []
                        for key, value in data.items():
                            result = self.select([col_to_check], {col_to_check: key})
                            if not result:
                                data_to_write.append(value)
                    else:
                        if isinstance(data, list):
                            data_to_write = data
                        else:
                            data_to_write = list(data.values())

                    if data_to_write:
                        joined_keys = '({})'.format(','.join([key.replace("'",
                                                                          "''") for key in data_to_write[0].keys()]))

                        values_joined_list = []
                        items_values = (item_data.values() for item_data in data_to_write)
                        for item_values in items_values:
                            joined_item_values = '({})'.format(','.join(["N'" + str(value).replace("'", "''") + "'"
                                                                         for value in item_values]))
                            values_joined_list.append(joined_item_values)

                        if 'date' in data_to_write[0]:
                            if not data_to_write[0].get('date'):
                                query = text("INSERT INTO {} {} VALUES {}".format(table_name,
                                                                                  '(id)', "('{}')".format(
                                                                                   data_to_write[0].get('id'))))
                            else:
                                joined_values = ','.join(values_joined_list)
                                query = text("INSERT INTO {} {} VALUES {}".format(table_name, joined_keys,
                                                                                  joined_values))
                        else:
                            joined_values = ','.join(values_joined_list)
                            query = text("INSERT INTO {} {} VALUES {}".format(table_name, joined_keys, joined_values))
                connection.execute(query)
                return True
            except:
                self.logger.write(40, 'Can\'t write data to the table. - \n\n{}'.format(format_exc()))
                return False

    def update(self, data, filter_parameters: Dict = {}, query_string: str = '', table_name: str = '') -> bool:
        """Makes UPDATE query to update data in database.

        @:param self:
        @:param data: data to update
        @:param filter_parameters: filters for WHERE query
        @:param query_string: custom query string
        @:param table_name: destination table name
        @:type self: DatabaseHelper
        @:type data: dict
        @:type filter_parameters: dict
        @:type query_string: str
        @:type table_name: str

        Makes UPDATE query.

        @:returns: True or False
        @:rtype: bool
        """

        with self.db_connection() as connection:
            try:
                if not query_string:
                    set_string = ','.join(['{}=\'{}\''.format(key.replace("'", "''"), str(value).replace("'", "''"))
                                           for key, value in data.items()])
                    query_string = 'UPDATE {} SET {}'.format(table_name, set_string)

                    if filter_parameters:
                        query_string += ' WHERE'

                        filter_parameters_counter = 1
                        filters_len = len(filter_parameters)

                        for filter_name, filter_parameter in filter_parameters.items():
                            if isinstance(filter_parameter, list):
                                filter_params = '(' + ','.join(
                                    ["N'" + str(value).replace("'", "''") + "'" for value in filter_parameter]) + ')'
                                column_match = ' {} IN {}'.format(filter_name, filter_params)
                            else:
                                column_match = ' {}={}'.format(filter_name,
                                                               "N'" + str(filter_parameter).replace("'", "''") + "'")
                            if filter_parameters_counter != filters_len:
                                column_match += ' AND'
                            query_string += column_match
                            filter_parameters_counter += 1

                connection.execute(text(query_string))
                return True
            except:
                self.logger.write(40, 'Can\'t update data in db. - \n\n{}'.format(format_exc()))
                return False

    def delete(self, filter_parameters: Dict = {}, query_string: str = '', table_name: str = '') -> bool:
        """Makes DELETE query.

        @:param self:
        @:param filter_parameters: filters for WHERE query
        @:param query_string: custom query string
        @:param table_name: destination table name
        @:type self: DatabaseHelper
        @:type filter_parameters: dict
        @:type query_string: str
        @:type table_name: str

        Makes DELETE query to delete data from database.

        @:returns: True or False
        @:rtype: bool
        """

        with self.db_connection() as connection:
            try:
                if not query_string:
                    query_string = 'DELETE FROM {}'.format(table_name)

                    if filter_parameters:
                        query_string += ' WHERE'

                        filter_parameters_counter = 1
                        filters_len = len(filter_parameters)

                        for filter_name, filter_parameter in filter_parameters.items():
                            if isinstance(filter_parameter, list):
                                filter_params = '(' + ','.join(
                                    ["N'" + str(value).replace("'", "''") + "'" for value in filter_parameter]) + ')'
                                column_match = ' {} IN {}'.format(filter_name, filter_params)
                            else:
                                column_match = ' {}={}'.format(filter_name,
                                                               "N'" + str(filter_parameter).replace("'", "''") + "'")
                            if filter_parameters_counter != filters_len:
                                column_match += ' AND'
                            query_string += column_match
                            filter_parameters_counter += 1

                connection.execute(text(query_string))
                return True
            except:
                self.logger.write(40, 'Can\'t update data in db. - \n\n{}'.format(format_exc()))
                return False

    @staticmethod
    def gen_batch_result(cursor, columns: List[str], fetchsize: int) -> List[Dict[str, Any]]:
        """Makes fetchmany from cursor to get results from query.

        @:param cursor: database cursor
        @:param columns: columns to zip with values
        @:param fetchsize: max amount of records which returns from request at once
        @:type cursor: Cursor
        @:type columns: list
        @:type fetchsize: int

        Makes fetchmany from cursor to get results -> if not results - break -> yields fetchmany results.

        @:returns: ASINs list
        @:rtype: list
        """

        while True:
            results = cursor.fetchmany(fetchsize)
            if not results:
                break
            else:
                modify_results = [dict(zip(columns, row)) for row in results]
                yield modify_results

    @contextmanager
    def db_connection(self):
        """Yields connection.

        Gets db credentials from config file -> creates connection -> yields engine.
        """

        server = self.db_credentials.get('server')
        database = self.db_credentials.get('database')
        username = self.db_credentials.get('username')
        password = self.db_credentials.get('password')

        # Setup connection parameters
        driver = '{/opt/microsoft/msodbcsql17/lib64/libmsodbcsql-17.5.so.2.1}'
        connection_string = 'DRIVER={};SERVER={};PORT=1433;DATABASE={};UID={};PWD={}'.format(driver, server, database,
                                                                                             username, password)

        params = parse.quote_plus(connection_string)
        engine = create_engine('mssql+pyodbc:///?odbc_connect=%s' % params, fast_executemany=True)

        connection = engine.connect()
        try:
            yield connection
        except exc.DatabaseError as err:
            self.logger.write(40, 'DatabaseError. - {}'.format(format_exc()))
            connection.execute('ROLLBACK')
            raise err
        finally:
            connection.close()



import configparser


def parse_config(part):
    config = configparser.ConfigParser()
    config.read('../config.ini')
    if part != 'ALL':
        config = config[part]
    return config


def chunkify(l, n):
    """Chunking list in n parts"""
    for i in range(0, len(l), n):
        yield l[i:i + n]






import requests
import logging
import time
from kupazavr_helpers.helpers import chunkify
from multiprocessing.pool import ThreadPool
from threading import RLock
from requests.exceptions import ProxyError, ConnectTimeout
import pandas as pd
import gc
from os import getenv
from typing import List, Union
from .user_agents import user_agents_list


class ProxySettings:
    """Proxy settings Instant"""

    def __init__(self, proxy_service=None, use_proxy=True, check_url='', proxy_country=None, per_proxy=None, country=None):
        self.proxy_service = proxy_service
        self.use_proxy = use_proxy
        self.check_url = check_url

        self.proxy_country = proxy_country
        self.per_proxy = per_proxy
        self.country = country
        if self.use_proxy and self.proxy_service == 'BEST' and not check_url:
            print('NO CHECK URL')
            exit()


class BadProxyError(Exception):
    def __init__(self):
        pass


class ProxyHelper:
    def __init__(self, logger, proxy_settings):
        self.proxy_settings = proxy_settings
        self.per_proxy = self.proxy_settings.per_proxy
        self.logger = logger
        self.user_agent = None
        self.proxy_service = self.proxy_settings.proxy_service
        self.country = self.proxy_settings.country
        self.check_url = self.proxy_settings.check_url
        self.use_proxy = self.proxy_settings.use_proxy
        self.proxy_pool = ThreadPool(100)
        if self.use_proxy:
            self.get_valid_proxies()
        self.lock = RLock()
        self.user_agents_list = user_agents_list

    def mark_proxy_as_failed(self, proxy):
        """
        Increase proxy error times and if it has more than 5, delete it
        :param proxy: proxy string with {ip}:{port} pattern
        :type proxy: str
        """
        logging.info(f'marking proxy {proxy}')
        self.proxies.at[proxy, 'previous_error_time'] = time.time()
        self.proxies.at[proxy, 'error_count'] += 1
        if int(self.proxies.at[proxy, 'error_count']) >= 5 and self.proxy_service == 'BEST':
            self.delete_proxy(proxy)

    def delete_proxy(self, proxy):
        """
        Deleting proxy from proxies dataframe
        :param proxy: proxy string with {ip}:{port} pattern
        :type proxy: str
        """
        logging.info('removing proxy {}'.format(proxy))
        self.proxies = self.proxies.drop(index=proxy)
        if len(self.proxies) == 0:
            logging.info('proxy list is empty, getting new proxies')
            self.get_valid_proxies()

    def exception_decorator(self, func):
        """
        Decorator for error tracking
        """

        def wrapper(*args, proxy, **kwargs):
            try:
                request_start_time = time.time()
                result = func(proxy, *args, **kwargs)
                request_time = time.time() - request_start_time
                with self.lock:
                    # proxy can be deleted from dataframe by another thread, so we need to check if it in still
                    if proxy and proxy in self.proxies.index:
                        self.proxies.at[proxy, 'previous_request_time'] = request_time
                with self.lock:
                    if proxy and proxy in self.proxies.index:
                        # decrease on_work variable for limiting proxy usage for 5 threads only
                        self.proxies.at[proxy, 'on_work'] -= 1
                return result
            except Exception as e:
                with self.lock:
                    if proxy and proxy in self.proxies.index:
                        self.proxies.at[proxy, 'on_work'] -= 1
                        if isinstance(e, (ProxyError, ConnectTimeout, BadProxyError)):
                            # marking proxy only if it error by proxy errors
                            self.mark_proxy_as_failed(proxy)
                raise

        return wrapper

    def get_proxy(self):
        """
        Getting best free proxy
        :return valid to use proxy
        :rtype: str
        """
        attempt = 0
        while True:
            try:
                self.sort_proxies()
                # getting proxies only which used less than 5 threads and get error more than 1 min ago
                with self.lock:
                    mask = ((time.time() - self.proxies['previous_error_time']) > 60) & (
                            self.proxies['on_work'] < self.per_proxy)
                    chosen_proxy = mask.idxmax()

                    self.proxies.at[chosen_proxy, 'on_work'] += 1
                    break
            except IndexError:
                # if we have no valid proxy. need to wait when it gets free
                logging.debug(f"Can\'t get proxy on {attempt} retry")
                time.sleep(0.5)
                attempt += 1
        return chosen_proxy

    def sort_proxies(self):
        """Sort proxies by best response time"""
        self.proxies = self.proxies.sort_values('on_work').sort_values('previous_request_time')

    def get_valid_proxies(self):
        """
        Proxy validation check
        """
        # loads all proxies from chosen proxy service
        proxies_list = self.get_proxies_list()
        chunks = list(chunkify(proxies_list, 500))
        valid_proxies = {}
        for n, chunk in enumerate(chunks):
            logging.info('checking {}/{} proxy batch for {}'.format(n + 1, len(chunks), self.check_url))
            # Check proxy batch
            checked_proxies = list(self.proxy_pool.map(self.check_proxy, chunk))
            gc.collect()
            # add primary proxy parameters
            valid_batch_proxies = {proxy['address']: {'previous_request_time': proxy['request_time'],
                                                      'error_count': 0,
                                                      'on_work': 0,
                                                      'previous_error_time': 0} for proxy in checked_proxies if
                                   proxy['is_valid']}
            valid_proxies.update(valid_batch_proxies)
            del checked_proxies
        # convert proxy dict into dataframe
        self.proxies = pd.DataFrame.from_dict(valid_proxies, orient='index')
        logging.info('proxy checking finished, found {} valid proxies'.format(len(self.proxies)))
        if self.proxies.empty:
            self.logger.write(40, 'NO PROXIES FOUND')
            return
        self.sort_proxies()

    def check_proxy(self, proxy):
        '''
        Check proxy on response time
        :param proxy: proxy string with {ip}:{port} pattern
        :type proxy: str
        :return: proxy primary data
        :rtype: dict
        '''
        try:
            if self.proxy_service == 'LOCAL':
                return {'address': proxy, 'is_valid': True, 'request_time': 0}
            t1 = time.time()
            headers = {'user-agent': self.get_user_agent()}
            response = requests.get(self.check_url, proxies={
                "http": proxy,
                "https": proxy,
            }, stream=True, timeout=10, headers=headers)
            for _ in response.iter_content(1024, decode_unicode=True):
                # if request time longer than 30 sec must stop request
                if time.time() - t1 > 30:
                    raise Exception
            request_time = time.time() - t1
            return {'address': proxy, 'is_valid': True, 'request_time': request_time}
        except:

            return {'address': proxy, 'is_valid': False}

    def get_proxies_list(self):
        '''Getting proxies from used proxy service'''
        if self.proxy_service == 'BEST':
            best_proxy_key = getenv('BEST_PROXY')
            if not best_proxy_key:
                self.logger.write(40, 'NO BEST PROXY KEY IN ENV')
                exit()
            proxies_list = self.load_proxies_list(best_proxy_key)
            if len(proxies_list) == 0:
                logging.error('get 0 proxies, get them from hardcoded for {}'.format(self.check_url))
        else:
            local_proxy = getenv('LOCAL_PROXY')
            if not local_proxy:
                self.logger.write(40, 'NO LOCAL PROXY IN ENV')
                exit()
            proxies_list = [local_proxy]
        return proxies_list

    def load_proxies_list(self, key):
        '''Request to proxy service'''
        proxies_list = []
        try:
            country = '&country={}'.format(self.country) if self.country else ''
            api_url = f'https://api.best-proxies.ru/proxylist.txt?key={key}&type=http,https&limit=0' + country
            r = requests.get(api_url, timeout=60)
            if r.status_code == 200:
                proxies_list = r.text.split('\r\n')
        except:
            pass
        return proxies_list

    def get_user_agent(self) -> Union[str, List]:
        """Rotates user agent.

        @:param self:
        @:type self: Downloader

        If user agent not None - appends user agent to the list -> gets first user agent from list.

        @:returns: user agent
        @:rtype: str
        """
        try:
            if self.user_agent:
                self.user_agents_list.append(self.user_agent)
            self.user_agent = self.user_agents_list.pop(0)
        except:
            return ''
        return self.user_agent



import base64
from urllib.parse import urlunparse, unquote
from urllib.request import _parse_proxy


def set_proxy(request, url):
    """
    Set proxy ip to Request Meta and put Proxy Authorization to request headers

    :param request: scrapy Request object
    :param url: proxy url
    :return:
    """
    credentials, proxy_url = parse_proxy(url)
    request.meta['proxy'] = proxy_url
    if credentials:
        request.headers['Proxy-Authorization'] = 'Basic ' + credentials.decode('utf-8')


def set_proxy_for_url(url):
    """
    Parse proxy url and set Proxy Authorization to headers dict

    :param url: proxy url
    :return: dict with Proxy Authorization header and raw proxy ip
    """
    headers = {}
    credentials, proxy_url = parse_proxy(url)
    if credentials:
        headers['Proxy-Authorization'] = 'Basic ' + credentials.decode('utf-8')

    return headers, proxy_url


def parse_proxy(url, _type='http'):
    """
    Proxy url parser.

    :param url: proxy url
    :param _type: url schema type
    :return: proxy credentials and proxy ip address
    """
    proxy_type, user, password, hostport = _parse_proxy(url)
    proxy_url = urlunparse((proxy_type or _type, hostport, '', '', '', ''))

    if user and password:
        user_password = f'{unquote(user)}:{unquote(password)}'
        credentials = base64.b64encode(user_password.encode('utf-8')).strip()
    else:
        credentials = None

    return credentials, proxy_url